/*JS*********************************************************************
*
*    Program : TESTSORT
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Test program for "gensort.h" sorting routines.
*
*              Copyright (C) 1994-2013 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*********     INCLUDES                                         *********/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#ifndef PAYLOAD
/* ***  Sorting an integer array  *** */
typedef unsigned long sorttype_t;

# define KEY(a)          a
# define FILL(a)
# define CHECKFILL(a)    1
#else
/* ***  Sorting a structure with some 'filling'  *** */
typedef struct {
  unsigned long Key;
  unsigned long Fill[PAYLOAD];

} sorttype_t;

# define KEY(a)          (a).Key
# define FILL(a)         dofill(&(a))
# define CHECKFILL(a)    checkfill(&(a))
#endif

/*  For generated mergesort  */
typedef struct llist_t llist_t;
struct llist_t {
  llist_t *Next;
  unsigned long Key;
};

#define Prototype
/*********     PROTOTYPES                                       *********/
static int               checkSort(sorttype_t *vector, sorttype_t *array,
				   unsigned long n, char *descript,
				   void (*routine)(sorttype_t *array, unsigned long n));
static int               checkMergeSort(sorttype_t *vector, llist_t *list,
					unsigned long n, const char *descript,
					llist_t *(*routine)(llist_t *list),
					int mode);

#ifdef PAYLOAD
static void              dofill(sorttype_t *a);
static int               checkfill(sorttype_t *a);
#endif

static void              quicksort(sorttype_t *array, unsigned long n);
static void              quicksortr(sorttype_t *array, unsigned long n);
static void              quicksortp(sorttype_t *array, unsigned long n);
static void              quicksortpr(sorttype_t *array, unsigned long n);
static void              heapsort(sorttype_t *array, unsigned long n);
static llist_t          *mergesort(llist_t *list);

static llist_t          *merge_McCarthy(llist_t *p, llist_t *q);
static llist_t          *mergesort_McCarthy(llist_t *head);

static llist_t          *mergesort_nonatural(llist_t *list);
static llist_t          *mergesort_nonlocal(llist_t *list);
static void              llquicksort(llist_t **array, unsigned long n);

static void              wrapqsort(sorttype_t *array, unsigned long n);
static int               longcmp(const void *a, const void *b);

/*********     GLOBAL/ STATIC VARIABLES                         *********/
static unsigned long QuickLimit = 0;

/*JS*********************************************************************
*
*************************************************************************/

int main(int argc, char **argv)

/************************************************************************/
{
  sorttype_t *vector, *array;
  llist_t *list;
  unsigned long mode, i, k, n, n2, mask;

  /*   Drop program name   */
  argv++; argc--;

  /*   Preset mode   */
  mode = 1;

  mask = -1;

  /*   Get arguments       */
  for(i = 0 ; argc-- > 0 ; argv++) {
    if(*argv) {
      switch(++i) {
      default:
	fprintf(stderr, "ERROR: To many arguments -- Break.\n");
	goto help;

      case 1: /*  "n"     */
	n = atoi(*argv);
	break;

      case 2: /*   "mode" */
	mode = atoi(*argv);
	break;

      case 3: /*   "seed" */
	srand( atoi(*argv) );
	break;

      case 4: /*   "limit" */
	QuickLimit = atoi(*argv);
	break;

      case 5: /*   "mask"  */
	mask = atoi(*argv);
	break;
      }
    }
  }

  /*   Less than one arguments?    */
  if(i < 1)
    goto help;

  /*   Get temporary memory for vector   */
  if((vector = (sorttype_t *)malloc(n * (2 * sizeof(*vector) + sizeof(*list)))) == NULL) {
    fprintf(stderr, "ERROR: Not enough memory (%ld Bytes needed) -- Break.\n",
	    2 * n * sizeof(*vector));
    exit(30);
  }
  array = vector + n;
  list = (llist_t *)&array[n];

  switch(mode) {
  default:
    printf("Unknown mode %ld -- Break.\n", mode);
    goto help;

  case 2:  /*  random but many (10) values are equal   */
    printf("Building random vector with all values appearing 10 times, length %ld...\n", n);
    n2 = n / 10;
    for(i = 0 ; i < n ; i++) {
      k = rand() % n2;
      k = abs(k);
      KEY(vector[i]) = k;
      FILL(vector[i]);
    }
    break;

  case 1:  /*   random   */
    printf("Building up random vector of length %ld....\n", n);
    n2 = n;
    for(i = 0 ; i < n ; i++) {
      k = rand() % n2;
      k = abs(k);
      KEY(vector[i]) = k;
      FILL(vector[i]);
    }
    break;

  case 3:  /*  ordered   */
    /*   Fill vector with random values    */
    printf("Building up ordered vector of length %ld.\n", n);
    for(i = 0 ; i < n ; i++) {
      KEY(vector[i]) = i;
      FILL(vector[i]);
    }
    break;

  case 4: /*  invers   */
    /*   Fill vector with random values    */
    printf("Building up invers ordered vector of length %ld....\n", n);
    for(i = 0 ; i < n ; i++) {
      KEY(vector[i]) = n - i;
      FILL(vector[i]);
    }
    break;
  }

  printf("Testing array sort routines:\n");

  if((mask & 1) && checkSort(vector, array, n, "Generated quicksort", quicksort))
    goto ende;

  if((mask & 2) && checkSort(vector, array, n, "Generated recursive quicksort", quicksortr))
    goto ende;

  if((mask & 4) && checkSort(vector, array, n, "Generated quicksort with pointers", quicksortp))
    goto ende;

  if((mask & 8) && checkSort(vector, array, n, "Generated recursive quicksort with pointers",
			     quicksortpr))
    goto ende;

  if((mask & 256) && checkSort(vector, array, n, "Generated heapsort", heapsort))
    goto ende;

  if((mask & 1024) && checkSort(vector, array, n, "Standard library quicksort", wrapqsort))
    goto ende;

  /* *********************************************************** */
  printf("Testing linked list sort routines:\n");

  if((mask & 4096) && checkMergeSort(vector, list, n, "Generated mergesort",
				     mergesort, 0))
    goto ende;

  if((mask & 4096) && checkMergeSort(vector, list, n, "Generated mergesort without natural runs",
				     mergesort_nonatural, 0))
    goto ende;

  if((mask & 4096) && checkMergeSort(vector, list, n, "Generated mergesort McCarthy algorithm",
				     mergesort_McCarthy, 0))
    goto ende;

  if((mask & 8192) && checkMergeSort(vector, list, n, "Generated non-local mergesort",
				     mergesort_nonlocal, 1))
    goto ende;

#if 0
  printf("The sorted vector:\n");
  for(i = 0 ; i < n ; i++)
    printf("%d%c", array[i], (i % 20) == 19 ? '\n' : ' ');
  putchar('\n');
#endif

ende:
  free(vector);

  return 0;
help:
  printf("Usage: testsort <n> [<mode> [<seed> [<limit>]]]\n\
 Tests the different types of sorting routines generated by \"gensort.h\".\n\
 Integer arrays of size <n> are sorted according to size. <limit> is the\n\
 threshhold for the array size below of which quicksort uses straight\n\
 insertion.\n\
 Mode: 1: random vector (Default).\n\
       2: random vector so that every value appears 10 times.\n\
       3: ordered vector.\n\
       4: invers ordered vector.\n");
  return 10;
}

/*JS*********************************************************************
*
*************************************************************************/

static int checkSort(sorttype_t *vector, sorttype_t *array, unsigned long n,
		     char *descript,
		     void (*routine)(sorttype_t *array, unsigned long n))

/************************************************************************/
{
  time_t tim;
  unsigned long i;

  /*  Make temporary copy of vector   */
  memcpy(array, vector, n * sizeof(vector[0]));

  printf("  %-40s...", descript); fflush(stdout);

  tim = clock();
  (*routine)(array, n);
  printf("%g seconds needed.", (clock() - tim) / (double) CLOCKS_PER_SEC);

  /*   Check sorting  */
  if(!CHECKFILL(array[0])) {
    printf("*** Rearranging invalid for a[0]!\n");
    goto error;
  }

  for(i = 1 ; i < n ; i++) {
    if(!CHECKFILL(array[i])) {
      printf("*** Rearranging invalid for a[%ld]!\n", i);
      goto error;
    }
    if(KEY(array[i - 1]) > KEY(array[i])) {
      printf("*** Sorting invalid: a[%ld] = %ld > a[%ld] = %ld!\n",
	     i - 1, KEY(array[i - 1]), i, KEY(array[i]));
      goto error;
    }
  }

  printf(" Sorting ok.\n");

  return 0;
error:
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

static int checkMergeSort(sorttype_t *vector, llist_t *list, unsigned long n,
			  const char *descript,
			  llist_t *(*routine)(llist_t *list), int mode)

/************************************************************************/
{
  llist_t *curr;
  time_t tim;
  unsigned long i, cnt;

  for(i = 0 ; i < n ; ++i) {
    list[i].Next = &list[i + 1];
    list[i].Key = KEY(vector[i]);
  }
  list[n - 1].Next = NULL;

  printf("  %-40s...", descript); fflush(stdout);
  tim = clock();

  curr = routine(list);

  printf("%g seconds needed.", (clock() - tim) / (double) CLOCKS_PER_SEC);

  for(cnt = 0 ; curr != NULL ; curr = curr->Next, ++cnt) {
    if(curr->Next != NULL && curr->Next->Key < curr->Key) {
      printf("*** Sorting invalid: %ld > %ld!\n",
	     curr->Key, curr->Next->Key);
      goto error;
    }
  }

  if(cnt != n) {
    printf("Got only %ld != %ld elements!\n", cnt, n);
    goto error;
  }

  printf(" Sorting ok.\n");

  if(mode & 1) {
    llist_t **array;

    if((array = (llist_t **)malloc(n * sizeof(*array))) == NULL)
      goto error;

    /*  For comparison, sort it via quicksort as a list  */
    for(i = 0 ; i < n ; ++i) {
      list[i].Next = &list[i + 1];
      list[i].Key = KEY(vector[i]);

      array[i] = list + i;
    }
    list[n - 1].Next = NULL;

    printf("  %-40s...", "Quicksort over list pointer array");
    fflush(stdout);

    tim = clock();

    llquicksort(array, n);

    printf("%g seconds needed.\n", (clock() - tim) / (double) CLOCKS_PER_SEC);

    free(array);
  }

  return 0;
 error:
  return -1;
}

#ifdef PAYLOAD
/*JS*********************************************************************
*
*************************************************************************/

static void dofill(sorttype_t *a)

/************************************************************************/
{
  unsigned long key;
  unsigned int i;

  key = a->Key;
  for(i = 0 ; i < PAYLOAD ; ++i)
    a->Fill[i] = key;
}

/*JS*********************************************************************
*
*************************************************************************/

static int checkfill(sorttype_t *a)

/************************************************************************/
{
  unsigned long key;
  unsigned int i;

  key = a->Key;
  for(i = 0 ; i < PAYLOAD ; ++i)
    if(a->Fill[i] != key) return(0);

  return(1);
}
#endif

/* ***  Generate normal QUICKSORT  *** */
#define GENSORT_NAME                 quicksort

#define GENSORT_LIMIT                QuickLimit
#define GENSORT_NO_RECURSIVE

#define GENSORT_TYPE                 sorttype_t
#define GENSORT_KEYTYPE              unsigned long
#define GENSORT_GETKEY(a)            KEY(a)
#define GENSORT_COMPAREKEYS(k1, k2)   k1 < k2

#define GENSORT_STATIC
#include "gensort.h"

/* ***  Generate recursive QUICKSORT  *** */
#define GENSORT_NAME                 quicksortr

#define GENSORT_LIMIT                QuickLimit
#define GENSORT_RECURSIVE

#define GENSORT_TYPE                 sorttype_t
#define GENSORT_KEYTYPE              unsigned long
#define GENSORT_GETKEY(a)            KEY(a)
#define GENSORT_COMPAREKEYS(k1, k2)   k1 < k2

#define GENSORT_STATIC
#include "gensort.h"

/* ***  Generate QUICKSORT with pointers  *** */
#define GENSORT_NAME                 quicksortp

#define GENSORT_LIMIT                QuickLimit
#define GENSORT_USEPOINTERS
#define GENSORT_NO_RECURSIVE

#define GENSORT_TYPE                 sorttype_t
#define GENSORT_KEYTYPE              unsigned long
#define GENSORT_GETKEY(a)            KEY(a)
#define GENSORT_COMPAREKEYS(k1, k2)   k1 < k2

#define GENSORT_STATIC
#include "gensort.h"

/* ***  Generate recursive QUICKSORT with pointers  *** */
#define GENSORT_NAME                 quicksortpr

#define GENSORT_LIMIT                QuickLimit
#define GENSORT_RECURSIVE

#define GENSORT_TYPE                 sorttype_t
#define GENSORT_KEYTYPE              unsigned long
#define GENSORT_GETKEY(a)            KEY(a)
#define GENSORT_COMPAREKEYS(k1, k2)   k1 < k2

#define GENSORT_STATIC
#include "gensort.h"

#undef GENSORT_USEPOINTERS

/* ***  Generate HEAPSORT  *** */
#define GENSORT_NAME                 heapsort

#define GENSORT_HEAPSORT

#define GENSORT_TYPE                 sorttype_t
#define GENSORT_KEYTYPE              unsigned long
#define GENSORT_GETKEY(a)            KEY(a)
#define GENSORT_COMPAREKEYS(k1, k2)   k1 < k2

#define GENSORT_STATIC
#include "gensort.h"

/* ************************************************************************ */

/* ***  Generate MERGESORT  *** */
#define GENSORT_NAME                 mergesort

#define GENSORT_MERGESORT
#define GENSORT_GETNEXT(a)           a->Next
#define GENSORT_SETNEXT(a, b)        a->Next = b

#define GENSORT_TYPE                 llist_t
#define GENSORT_KEYTYPE              unsigned long
#define GENSORT_GETKEY(a)            a.Key
#define GENSORT_COMPAREKEYS(k1, k2)  k1 < k2

#define GENSORT_STATIC
#include "gensort.h"

/* ***  MERGESORT with McCarthy algorithm  *** */

/*  The next two routines are courtesy of Sosman <esosman@acm-dot-org.invalid>  */

#define IMPROVED_MCCARTHY

/*
 * Internal routine to merge two non-empty sorted lists.  If elements
 * of the two lists compare equal, those from the `p' list will precede
 * those from the `q' list in the merged output.
 */
static llist_t *merge_McCarthy(llist_t *p, llist_t *q)
{
  llist_t head, *tail;

  tail = &head;
  for (;;) {
    if(p->Key <= q->Key) {
      tail->Next = p;
      tail = p;
      if((p = p->Next) == NULL) {
	tail->Next = q;
	break;
      }
    } else {
      tail->Next = q;
      tail = q;
      if((q = q->Next) == NULL) {
	tail->Next = p;
	break;
      }
    }
  }

  return head.Next;
}

/*
 * rearranges the links of a singly-linked NULL-terminated
 * list so they traverse the items in order, and returns a pointer to the first
 * item in the rearranged list. The sort is stable.
 */
static llist_t *mergesort_McCarthy(llist_t *head)
{
  llist_t *alist[sizeof(unsigned long) * 8];
#ifdef IMPROVED_MCCARTHY
  llist_t *blist[sizeof(unsigned long) * 8];
#endif
  llist_t *p, *q;
  unsigned int bits, maxbits;

  maxbits = 0;
  alist[0] = NULL;
#ifdef IMPROVED_MCCARTHY
  blist[0] = NULL;
#endif
  while((p = head) != NULL) {
    if((q = p->Next) == NULL)
      break;

    head = q->Next;
    if(p->Key < q->Key) {
      q->Next = NULL;
    } else {
      q->Next = p;
      p->Next = NULL;
      p = q;
    }

#ifdef IMPROVED_MCCARTHY
    for(bits = 0 ; ; p = q) {
      if(alist[bits] == NULL) {
	alist[bits] = p;
	break;
      }
      if(blist[bits] == NULL) {
	blist[bits] = p;
	break;
      }
      q = merge_McCarthy(alist[bits], blist[bits]);
      alist[bits] = p;
      blist[bits] = NULL;
      if(++bits > maxbits) {
	maxbits = bits;
	alist[bits] = q;
	blist[bits] = NULL;
	break;
      }
    }
#else
    for(bits = 0 ; (q = alist[bits]) != NULL ; ) {
      alist[bits] = NULL;
      p = merge_McCarthy(q, p);
      if(++bits > maxbits) {
	maxbits = bits;
	break;
      }
    }

    alist[bits] = p;
#endif
  }

  for(bits = 0 ; bits <= maxbits ; ++bits) {
#ifdef IMPROVED_MCCARTHY
    if((p = blist[bits]) != NULL)
      head = (head == NULL) ? p : merge_McCarthy(p, head);
#endif

    if((p = alist[bits]) != NULL)
      head = (head == NULL) ? p : merge_McCarthy(p, head);
  }

  return head;
}

/* ***  Generate MERGESORT without natural runs  *** */
#define GENSORT_NAME                 mergesort_nonatural

#define GENSORT_MERGESORT
#define GENSORT_MERGESORT_NONATURAL

#define GENSORT_GETNEXT(a)           a->Next
#define GENSORT_SETNEXT(a, b)        a->Next = b

#define GENSORT_TYPE                 llist_t
#define GENSORT_KEYTYPE              unsigned long
#define GENSORT_GETKEY(a)            a.Key
#define GENSORT_COMPAREKEYS(k1, k2)  k1 < k2

#define GENSORT_STATIC
#include "gensort.h"

/* ***  Generate MERGESORT (non-local) *** */
#define GENSORT_NAME                 mergesort_nonlocal

#define GENSORT_MERGESORT
#define GENSORT_MERGESORT_NONLOCAL
#define GENSORT_GETNEXT(a)           a->Next
#define GENSORT_SETNEXT(a, b)        a->Next = b

#define GENSORT_TYPE                 llist_t
#define GENSORT_KEYTYPE              unsigned long
#define GENSORT_GETKEY(a)            a.Key
#define GENSORT_COMPAREKEYS(k1, k2)  k1 < k2

#define GENSORT_STATIC
#include "gensort.h"

/* ***  Generate normal QUICKSORT for linked list *** */
#define GENSORT_NAME                 llquicksort

#define GENSORT_TYPE                 llist_t *
#define GENSORT_KEYTYPE              unsigned long
#define GENSORT_GETKEY(a)            a->Key
#define GENSORT_COMPAREKEYS(k1, k2)  k1 < k2

#define GENSORT_STATIC
#include "gensort.h"

/* ***  Wrapper for standard library qsort  *** */
static void wrapqsort(sorttype_t *array, unsigned long n)
{
  qsort(array, n, sizeof(*array), longcmp);
}

static int longcmp(const void *a, const void *b)
{
  return(KEY(*(sorttype_t *)a) < KEY(*(sorttype_t *)b) ? -1 :
	 KEY(*(sorttype_t *)a) > KEY(*(sorttype_t *)b) ? 1 : 0);
}
